package build

import (
	"bufio"
	"os"
	"path/filepath"
	"strings"

	"git.akyoto.dev/cli/q/src/asm"
	"git.akyoto.dev/cli/q/src/directory"
	"git.akyoto.dev/cli/q/src/elf"
	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/cli/q/src/linux"
	"git.akyoto.dev/cli/q/src/log"
	"git.akyoto.dev/cli/q/src/register"
)

// Build describes a compiler build.
type Build struct {
	Directory       string
	Verbose         bool
	WriteExecutable bool
}

// New creates a new build.
func New(directory string) *Build {
	return &Build{
		Directory:       directory,
		WriteExecutable: true,
	}
}

var (
	hello = []byte("Hello\n")
	world = []byte("World\n")
)

// Run parses the input files and generates an executable file.
func (build *Build) Run() error {
	// err := build.Compile()

	// if err != nil {
	// 	return err
	// }

	a := asm.Assembler{
		Instructions: make([]asm.Instruction, 0, 8),
		Verbose:      build.Verbose,
	}

	a.MoveRegisterNumber(register.Syscall0, linux.Write)
	a.MoveRegisterNumber(register.Syscall1, 1)
	a.MoveRegisterData(register.Syscall2, hello)
	a.MoveRegisterNumber(register.Syscall3, uint64(len(hello)))
	a.Syscall()

	a.MoveRegisterNumber(register.Syscall0, linux.Write)
	a.MoveRegisterNumber(register.Syscall1, 1)
	a.MoveRegisterData(register.Syscall2, world)
	a.MoveRegisterNumber(register.Syscall3, uint64(len(world)))
	a.Syscall()

	a.MoveRegisterNumber(register.Syscall0, linux.Exit)
	a.MoveRegisterNumber(register.Syscall1, 0)
	a.Syscall()

	code, data := a.Finalize()

	if !build.WriteExecutable {
		return nil
	}

	return writeToDisk(build.Executable(), code, data)
}

// Compile compiles all the functions.
func (build *Build) Compile() error {
	stat, err := os.Stat(build.Directory)

	if err != nil {
		return err
	}

	if !stat.IsDir() {
		return &errors.InvalidDirectory{Path: build.Directory}
	}

	directory.Walk(build.Directory, func(file string) {
		if !strings.HasSuffix(file, ".q") {
			return
		}

		log.Info.Println(file)
	})

	return nil
}

// Executable returns the path to the executable.
func (build *Build) Executable() string {
	return filepath.Join(build.Directory, filepath.Base(build.Directory))
}

// writeToDisk writes the executable file to disk.
func writeToDisk(filePath string, code []byte, data []byte) error {
	file, err := os.Create(filePath)

	if err != nil {
		return err
	}

	buffer := bufio.NewWriter(file)
	executable := elf.New(code, data)
	executable.Write(buffer)
	buffer.Flush()

	err = file.Close()

	if err != nil {
		return err
	}

	return os.Chmod(filePath, 0755)
}
